@@ -14,13 +14,15 @@
 // well.
 
 use std::collections::HashMap;
-use std::ffi::{OsString, OsStr};
+use std::ffi::{OsStr, OsString};
+use std::fs::{File, FileType};
+use std::io::{self, BufRead};
 use std::path::{Path, PathBuf};
 use std::sync::{Arc, RwLock};
 
 use gitignore::{self, Gitignore, GitignoreBuilder};
-use pathutil::{is_hidden, strip_prefix};
 use overrides::{self, Override};
+use pathutil::{is_hidden, strip_prefix};
 use types::{self, Types};
 use walk::DirEntry;
 use {Error, Match, PartialErrorBuilder};
@@ -76,6 +78,9 @@ struct IgnoreOptions {
     git_exclude: bool,
     /// Whether to ignore files case insensitively
     ignore_case_insensitive: bool,
+    /// Whether a git repository must be present in order to apply any
+    /// git-related ignore rules.
+    require_git: bool,
 }
 
 /// Ignore is a matcher useful for recursively walking one or more directories.
@@ -197,12 +202,11 @@ impl Ignore {
             errs.maybe_push(err);
             igtmp.is_absolute_parent = true;
             igtmp.absolute_base = Some(absolute_base.clone());
-            igtmp.has_git =
-                if self.0.opts.git_ignore {
-                    parent.join(".git").exists()
-                } else {
-                    false
-                };
+            igtmp.has_git = if self.0.opts.git_ignore {
+                parent.join(".git").exists()
+            } else {
+                false
+            };
             ig = Ignore(Arc::new(igtmp));
             compiled.insert(parent.as_os_str().to_os_string(), ig.clone());
         }
@@ -227,65 +231,70 @@ impl Ignore {
 
     /// Like add_child, but takes a full path and returns an IgnoreInner.
     fn add_child_path(&self, dir: &Path) -> (IgnoreInner, Option<Error>) {
+        let git_type = if self.0.opts.git_ignore || self.0.opts.git_exclude {
+            dir.join(".git").metadata().ok().map(|md| md.file_type())
+        } else {
+            None
+        };
+        let has_git = git_type.map(|_| true).unwrap_or(false);
+
         let mut errs = PartialErrorBuilder::default();
-        let custom_ig_matcher =
-            if self.0.custom_ignore_filenames.is_empty() {
-                Gitignore::empty()
-            } else {
-                let (m, err) =
-                    create_gitignore(
+        let custom_ig_matcher = if self.0.custom_ignore_filenames.is_empty() {
+            Gitignore::empty()
+        } else {
+            let (m, err) = create_gitignore(
+                &dir,
+                &dir,
+                &self.0.custom_ignore_filenames,
+                self.0.opts.ignore_case_insensitive,
+            );
+            errs.maybe_push(err);
+            m
+        };
+        let ig_matcher = if !self.0.opts.ignore {
+            Gitignore::empty()
+        } else {
+            let (m, err) = create_gitignore(
+                &dir,
+                &dir,
+                &[".ignore"],
+                self.0.opts.ignore_case_insensitive,
+            );
+            errs.maybe_push(err);
+            m
+        };
+        let gi_matcher = if !self.0.opts.git_ignore {
+            Gitignore::empty()
+        } else {
+            let (m, err) = create_gitignore(
+                &dir,
+                &dir,
+                &[".gitignore"],
+                self.0.opts.ignore_case_insensitive,
+            );
+            errs.maybe_push(err);
+            m
+        };
+        let gi_exclude_matcher = if !self.0.opts.git_exclude {
+            Gitignore::empty()
+        } else {
+            match resolve_git_commondir(dir, git_type) {
+                Ok(git_dir) => {
+                    let (m, err) = create_gitignore(
                         &dir,
-                        &self.0.custom_ignore_filenames,
+                        &git_dir,
+                        &["info/exclude"],
                         self.0.opts.ignore_case_insensitive,
                     );
-                errs.maybe_push(err);
-                m
-            };
-        let ig_matcher =
-            if !self.0.opts.ignore {
-                Gitignore::empty()
-            } else {
-                let (m, err) =
-                    create_gitignore(
-                        &dir,
-                        &[".ignore"],
-                        self.0.opts.ignore_case_insensitive,
-                    );
-                errs.maybe_push(err);
-                m
-            };
-        let gi_matcher =
-            if !self.0.opts.git_ignore {
-                Gitignore::empty()
-            } else {
-                let (m, err) =
-                    create_gitignore(
-                        &dir,
-                        &[".gitignore"],
-                        self.0.opts.ignore_case_insensitive,
-                    );
-                errs.maybe_push(err);
-                m
-            };
-        let gi_exclude_matcher =
-            if !self.0.opts.git_exclude {
-                Gitignore::empty()
-            } else {
-                let (m, err) =
-                    create_gitignore(
-                        &dir,
-                        &[".git/info/exclude"],
-                        self.0.opts.ignore_case_insensitive,
-                    );
-                errs.maybe_push(err);
-                m
-            };
-        let has_git =
-            if self.0.opts.git_ignore {
-                dir.join(".git").exists()
-            } else {
-                false
-            };
+                    errs.maybe_push(err);
+                    m
+                }
+                Err(err) => {
+                    errs.maybe_push(err);
+                    Gitignore::empty()
+                }
+            }
+        };
         let ig = IgnoreInner {
             compiled: self.0.compiled.clone(),
             dir: dir.to_path_buf(),
@@ -301,7 +310,7 @@ impl Ignore {
             git_global_matcher: self.0.git_global_matcher.clone(),
             git_ignore_matcher: gi_matcher,
             git_exclude_matcher: gi_exclude_matcher,
-            has_git: has_git,
+            has_git,
             opts: self.0.opts,
         };
         (ig, errs.into_error_option())
@@ -310,12 +319,16 @@ impl Ignore {
     /// Returns true if at least one type of ignore rule should be matched.
     fn has_any_ignore_rules(&self) -> bool {
         let opts = self.0.opts;
-        let has_custom_ignore_files = !self.0.custom_ignore_filenames.is_empty();
+        let has_custom_ignore_files =
+            !self.0.custom_ignore_filenames.is_empty();
         let has_explicit_ignores = !self.0.explicit_ignores.is_empty();
 
-        opts.ignore || opts.git_global || opts.git_ignore
-                    || opts.git_exclude || has_custom_ignore_files
-                    || has_explicit_ignores
+        opts.ignore
+            || opts.git_global
+            || opts.git_ignore
+            || opts.git_exclude
+            || has_custom_ignore_files
+            || has_explicit_ignores
     }
 
     /// Like `matched`, but works with a directory entry instead.
@@ -350,9 +363,11 @@ impl Ignore {
         // return that result immediately. Overrides have the highest
         // precedence.
         if !self.0.overrides.is_empty() {
-            let mat =
-                self.0.overrides.matched(path, is_dir)
-                    .map(IgnoreMatch::overrides);
+            let mat = self
+                .0
+                .overrides
+                .matched(path, is_dir)
+                .map(IgnoreMatch::overrides);
             if !mat.is_none() {
                 return mat;
             }
@@ -385,56 +400,72 @@ impl Ignore {
         path: &Path,
         is_dir: bool,
     ) -> Match<IgnoreMatch<'a>> {
-        let (mut m_custom_ignore, mut m_ignore, mut m_gi, mut m_gi_exclude, mut m_explicit) =
-            (Match::None, Match::None, Match::None, Match::None, Match::None);
-        let any_git = self.parents().any(|ig| ig.0.has_git);
+        let (
+            mut m_custom_ignore,
+            mut m_ignore,
+            mut m_gi,
+            mut m_gi_exclude,
+            mut m_explicit,
+        ) = (Match::None, Match::None, Match::None, Match::None, Match::None);
+        let any_git =
+            !self.0.opts.require_git || self.parents().any(|ig| ig.0.has_git);
         let mut saw_git = false;
         for ig in self.parents().take_while(|ig| !ig.0.is_absolute_parent) {
             if m_custom_ignore.is_none() {
                 m_custom_ignore =
-                    ig.0.custom_ignore_matcher.matched(path, is_dir)
-                      .map(IgnoreMatch::gitignore);
+                    ig.0.custom_ignore_matcher
+                        .matched(path, is_dir)
+                        .map(IgnoreMatch::gitignore);
             }
             if m_ignore.is_none() {
                 m_ignore =
-                    ig.0.ignore_matcher.matched(path, is_dir)
-                      .map(IgnoreMatch::gitignore);
+                    ig.0.ignore_matcher
+                        .matched(path, is_dir)
+                        .map(IgnoreMatch::gitignore);
             }
             if any_git && !saw_git && m_gi.is_none() {
                 m_gi =
-                    ig.0.git_ignore_matcher.matched(path, is_dir)
-                      .map(IgnoreMatch::gitignore);
+                    ig.0.git_ignore_matcher
+                        .matched(path, is_dir)
+                        .map(IgnoreMatch::gitignore);
             }
             if any_git && !saw_git && m_gi_exclude.is_none() {
                 m_gi_exclude =
-                    ig.0.git_exclude_matcher.matched(path, is_dir)
-                      .map(IgnoreMatch::gitignore);
+                    ig.0.git_exclude_matcher
+                        .matched(path, is_dir)
+                        .map(IgnoreMatch::gitignore);
             }
             saw_git = saw_git || ig.0.has_git;
         }
         if self.0.opts.parents {
             if let Some(abs_parent_path) = self.absolute_base() {
                 let path = abs_parent_path.join(path);
-                for ig in self.parents().skip_while(|ig|!ig.0.is_absolute_parent) {
+                for ig in
+                    self.parents().skip_while(|ig| !ig.0.is_absolute_parent)
+                {
                     if m_custom_ignore.is_none() {
                         m_custom_ignore =
-                            ig.0.custom_ignore_matcher.matched(&path, is_dir)
-                              .map(IgnoreMatch::gitignore);
+                            ig.0.custom_ignore_matcher
+                                .matched(&path, is_dir)
+                                .map(IgnoreMatch::gitignore);
                     }
                     if m_ignore.is_none() {
                         m_ignore =
-                            ig.0.ignore_matcher.matched(&path, is_dir)
-                              .map(IgnoreMatch::gitignore);
+                            ig.0.ignore_matcher
+                                .matched(&path, is_dir)
+                                .map(IgnoreMatch::gitignore);
                     }
                     if any_git && !saw_git && m_gi.is_none() {
                         m_gi =
-                            ig.0.git_ignore_matcher.matched(&path, is_dir)
-                              .map(IgnoreMatch::gitignore);
+                            ig.0.git_ignore_matcher
+                                .matched(&path, is_dir)
+                                .map(IgnoreMatch::gitignore);
                     }
                     if any_git && !saw_git && m_gi_exclude.is_none() {
                         m_gi_exclude =
-                            ig.0.git_exclude_matcher.matched(&path, is_dir)
-                              .map(IgnoreMatch::gitignore);
+                            ig.0.git_exclude_matcher
+                                .matched(&path, is_dir)
+                                .map(IgnoreMatch::gitignore);
                     }
                     saw_git = saw_git || ig.0.has_git;
                 }
@@ -446,16 +477,21 @@ impl Ignore {
             }
             m_explicit = gi.matched(&path, is_dir).map(IgnoreMatch::gitignore);
         }
-        let m_global =
-            if any_git {
-                self.0.git_global_matcher
-                    .matched(&path, is_dir)
-                    .map(IgnoreMatch::gitignore)
-            } else {
-                Match::None
-            };
+        let m_global = if any_git {
+            self.0
+                .git_global_matcher
+                .matched(&path, is_dir)
+                .map(IgnoreMatch::gitignore)
+        } else {
+            Match::None
+        };
 
-        m_custom_ignore.or(m_ignore).or(m_gi).or(m_gi_exclude).or(m_global).or(m_explicit)
+        m_custom_ignore
+            .or(m_ignore)
+            .or(m_gi)
+            .or(m_gi_exclude)
+            .or(m_global)
+            .or(m_explicit)
     }
 
     /// Returns an iterator over parent ignore matchers, including this one.
@@ -526,6 +562,7 @@ impl IgnoreBuilder {
                 git_ignore: true,
                 git_exclude: true,
                 ignore_case_insensitive: false,
+                require_git: true,
             },
         }
     }
@@ -535,20 +572,19 @@ impl IgnoreBuilder {
     /// The matcher returned won't match anything until ignore rules from
     /// directories are added to it.
     pub fn build(&self) -> Ignore {
-        let git_global_matcher =
-            if !self.opts.git_global {
-                Gitignore::empty()
-            } else {
-                let mut builder = GitignoreBuilder::new("");
-                builder
-                    .case_insensitive(self.opts.ignore_case_insensitive)
-                    .unwrap();
-                let (gi, err) = builder.build_global();
-                if let Some(err) = err {
-                    debug!("{}", err);
-                }
-                gi
-            };
+        let git_global_matcher = if !self.opts.git_global {
+            Gitignore::empty()
+        } else {
+            let mut builder = GitignoreBuilder::new("");
+            builder
+                .case_insensitive(self.opts.ignore_case_insensitive)
+                .unwrap();
+            let (gi, err) = builder.build_global();
+            if let Some(err) = err {
+                debug!("{}", err);
+            }
+            gi
+        };
 
         Ignore(Arc::new(IgnoreInner {
             compiled: Arc::new(RwLock::new(HashMap::new())),
@@ -559,7 +595,9 @@ impl IgnoreBuilder {
             is_absolute_parent: true,
             absolute_base: None,
             explicit_ignores: Arc::new(self.explicit_ignores.clone()),
-            custom_ignore_filenames: Arc::new(self.custom_ignore_filenames.clone()),
+            custom_ignore_filenames: Arc::new(
+                self.custom_ignore_filenames.clone(),
+            ),
             custom_ignore_matcher: Gitignore::empty(),
             ignore_matcher: Gitignore::empty(),
             git_global_matcher: Arc::new(git_global_matcher),
@@ -604,7 +642,7 @@ impl IgnoreBuilder {
     /// later names.
     pub fn add_custom_ignore_filename<S: AsRef<OsStr>>(
         &mut self,
-        file_name: S
+        file_name: S,
     ) -> &mut IgnoreBuilder {
         self.custom_ignore_filenames.push(file_name.as_ref().to_os_string());
         self
@@ -675,6 +713,16 @@ impl IgnoreBuilder {
         self
     }
 
+    /// Whether a git repository is required to apply git-related ignore
+    /// rules (global rules, .gitignore and local exclude rules).
+    ///
+    /// When disabled, git-related ignore rules are applied even when searching
+    /// outside a git repository.
+    pub fn require_git(&mut self, yes: bool) -> &mut IgnoreBuilder {
+        self.opts.require_git = yes;
+        self
+    }
+
     /// Process ignore files case insensitively
     ///
     /// This is disabled by default.
@@ -689,12 +737,15 @@ impl IgnoreBuilder {
 
 /// Creates a new gitignore matcher for the directory given.
 ///
-/// Ignore globs are extracted from each of the file names in `dir` in the
-/// order given (earlier names have lower precedence than later names).
+/// The matcher is meant to match files below `dir`.
+/// Ignore globs are extracted from each of the file names relative to
+/// `dir_for_ignorefile` in the order given (earlier names have lower
+/// precedence than later names).
 ///
 /// I/O errors are ignored.
 pub fn create_gitignore<T: AsRef<OsStr>>(
     dir: &Path,
+    dir_for_ignorefile: &Path,
     names: &[T],
     case_insensitive: bool,
 ) -> (Gitignore, Option<Error>) {
@@ -702,8 +753,22 @@ pub fn create_gitignore<T: AsRef<OsStr>>(
     let mut errs = PartialErrorBuilder::default();
     builder.case_insensitive(case_insensitive).unwrap();
     for name in names {
-        let gipath = dir.join(name.as_ref());
-        errs.maybe_push_ignore_io(builder.add(gipath));
+        let gipath = dir_for_ignorefile.join(name.as_ref());
+        // This check is not necessary, but is added for performance. Namely,
+        // a simple stat call checking for existence can often be just a bit
+        // quicker than actually trying to open a file. Since the number of
+        // directories without ignore files likely greatly exceeds the number
+        // with ignore files, this check generally makes sense.
+        //
+        // However, until demonstrated otherwise, we speculatively do not do
+        // this on Windows since Windows is notorious for having slow file
+        // system operations. Namely, it's not clear whether this analysis
+        // makes sense on Windows.
+        //
+        // For more details: https://github.com/BurntSushi/ripgrep/pull/1381
+        if cfg!(windows) || gipath.exists() {
+            errs.maybe_push_ignore_io(builder.add(gipath));
+        }
     }
     let gi = match builder.build() {
         Ok(gi) => gi,
@@ -715,6 +780,60 @@ pub fn create_gitignore<T: AsRef<OsStr>>(
     (gi, errs.into_error_option())
 }
 
+/// Find the GIT_COMMON_DIR for the given git worktree.
+///
+/// This is the directory that may contain a private ignore file
+/// "info/exclude". Unlike git, this function does *not* read environment
+/// variables GIT_DIR and GIT_COMMON_DIR, because it is not clear how to use
+/// them when multiple repositories are searched.
+///
+/// Some I/O errors are ignored.
+fn resolve_git_commondir(
+    dir: &Path,
+    git_type: Option<FileType>,
+) -> Result<PathBuf, Option<Error>> {
+    let git_dir_path = || dir.join(".git");
+    let git_dir = git_dir_path();
+    if !git_type.map_or(false, |ft| ft.is_file()) {
+        return Ok(git_dir);
+    }
+    let file = match File::open(git_dir) {
+        Ok(file) => io::BufReader::new(file),
+        Err(err) => {
+            return Err(Some(Error::Io(err).with_path(git_dir_path())));
+        }
+    };
+    let dot_git_line = match file.lines().next() {
+        Some(Ok(line)) => line,
+        Some(Err(err)) => {
+            return Err(Some(Error::Io(err).with_path(git_dir_path())));
+        }
+        None => return Err(None),
+    };
+    if !dot_git_line.starts_with("gitdir: ") {
+        return Err(None);
+    }
+    let real_git_dir = PathBuf::from(&dot_git_line["gitdir: ".len()..]);
+    let git_commondir_file = || real_git_dir.join("commondir");
+    let file = match File::open(git_commondir_file()) {
+        Ok(file) => io::BufReader::new(file),
+        Err(_) => return Err(None),
+    };
+    let commondir_line = match file.lines().next() {
+        Some(Ok(line)) => line,
+        Some(Err(err)) => {
+            return Err(Some(Error::Io(err).with_path(git_commondir_file())));
+        }
+        None => return Err(None),
+    };
+    let commondir_abs = if commondir_line.starts_with(".") {
+        real_git_dir.join(commondir_line) // relative commondir
+    } else {
+        PathBuf::from(commondir_line)
+    };
+    Ok(commondir_abs)
+}
+
 #[cfg(test)]
 mod tests {
     use std::fs::{self, File};
@@ -742,19 +861,19 @@ mod tests {
         }
     }
 
-    fn tmpdir(prefix: &str) -> TempDir {
+    fn tmpdir() -> TempDir {
         TempDir::new().unwrap()
     }
 
     #[test]
     fn explicit_ignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join("not-an-ignore"), "foo\n!bar");
 
         let (gi, err) = Gitignore::new(td.path().join("not-an-ignore"));
         assert!(err.is_none());
-        let (ig, err) = IgnoreBuilder::new()
-            .add_ignore(gi).build().add_child(td.path());
+        let (ig, err) =
+            IgnoreBuilder::new().add_ignore(gi).build().add_child(td.path());
         assert!(err.is_none());
         assert!(ig.matched("foo", false).is_ignore());
         assert!(ig.matched("bar", false).is_whitelist());
@@ -763,7 +882,7 @@ mod tests {
 
     #[test]
     fn git_exclude() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         mkdirp(td.path().join(".git/info"));
         wfile(td.path().join(".git/info/exclude"), "foo\n!bar");
 
@@ -776,7 +895,7 @@ mod tests {
 
     #[test]
     fn gitignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         mkdirp(td.path().join(".git"));
         wfile(td.path().join(".gitignore"), "foo\n!bar");
 
@@ -789,7 +908,7 @@ mod tests {
 
     #[test]
     fn gitignore_no_git() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".gitignore"), "foo\n!bar");
 
         let (ig, err) = IgnoreBuilder::new().build().add_child(td.path());
@@ -799,9 +918,24 @@ mod tests {
         assert!(ig.matched("baz", false).is_none());
     }
 
+    #[test]
+    fn gitignore_allowed_no_git() {
+        let td = tmpdir();
+        wfile(td.path().join(".gitignore"), "foo\n!bar");
+
+        let (ig, err) = IgnoreBuilder::new()
+            .require_git(false)
+            .build()
+            .add_child(td.path());
+        assert!(err.is_none());
+        assert!(ig.matched("foo", false).is_ignore());
+        assert!(ig.matched("bar", false).is_whitelist());
+        assert!(ig.matched("baz", false).is_none());
+    }
+
     #[test]
     fn ignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".ignore"), "foo\n!bar");
 
         let (ig, err) = IgnoreBuilder::new().build().add_child(td.path());
@@ -813,13 +947,14 @@ mod tests {
 
     #[test]
     fn custom_ignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         let custom_ignore = ".customignore";
         wfile(td.path().join(custom_ignore), "foo\n!bar");
 
         let (ig, err) = IgnoreBuilder::new()
             .add_custom_ignore_filename(custom_ignore)
-            .build().add_child(td.path());
+            .build()
+            .add_child(td.path());
         assert!(err.is_none());
         assert!(ig.matched("foo", false).is_ignore());
         assert!(ig.matched("bar", false).is_whitelist());
@@ -829,14 +964,15 @@ mod tests {
     // Tests that a custom ignore file will override an .ignore.
     #[test]
     fn custom_ignore_over_ignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         let custom_ignore = ".customignore";
         wfile(td.path().join(".ignore"), "foo");
         wfile(td.path().join(custom_ignore), "!foo");
 
         let (ig, err) = IgnoreBuilder::new()
             .add_custom_ignore_filename(custom_ignore)
-            .build().add_child(td.path());
+            .build()
+            .add_child(td.path());
         assert!(err.is_none());
         assert!(ig.matched("foo", false).is_whitelist());
     }
@@ -844,7 +980,7 @@ mod tests {
     // Tests that earlier custom ignore files have lower precedence than later.
     #[test]
     fn custom_ignore_precedence() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         let custom_ignore1 = ".customignore1";
         let custom_ignore2 = ".customignore2";
         wfile(td.path().join(custom_ignore1), "foo");
@@ -853,7 +989,8 @@ mod tests {
         let (ig, err) = IgnoreBuilder::new()
             .add_custom_ignore_filename(custom_ignore1)
             .add_custom_ignore_filename(custom_ignore2)
-            .build().add_child(td.path());
+            .build()
+            .add_child(td.path());
         assert!(err.is_none());
         assert!(ig.matched("foo", false).is_whitelist());
     }
@@ -861,7 +998,7 @@ mod tests {
     // Tests that an .ignore will override a .gitignore.
     #[test]
     fn ignore_over_gitignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".gitignore"), "foo");
         wfile(td.path().join(".ignore"), "!foo");
 
@@ -873,7 +1010,7 @@ mod tests {
     // Tests that exclude has lower precedent than both .ignore and .gitignore.
     #[test]
     fn exclude_lowest() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".gitignore"), "!foo");
         wfile(td.path().join(".ignore"), "!bar");
         mkdirp(td.path().join(".git/info"));
@@ -888,7 +1025,7 @@ mod tests {
 
     #[test]
     fn errored() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".gitignore"), "{foo");
 
         let (_, err) = IgnoreBuilder::new().build().add_child(td.path());
@@ -897,7 +1034,7 @@ mod tests {
 
     #[test]
     fn errored_both() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".gitignore"), "{foo");
         wfile(td.path().join(".ignore"), "{bar");
 
@@ -907,7 +1044,7 @@ mod tests {
 
     #[test]
     fn errored_partial() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         mkdirp(td.path().join(".git"));
         wfile(td.path().join(".gitignore"), "{foo\nbar");
 
@@ -918,7 +1055,7 @@ mod tests {
 
     #[test]
     fn errored_partial_and_ignore() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         wfile(td.path().join(".gitignore"), "{foo\nbar");
         wfile(td.path().join(".ignore"), "!bar");
 
@@ -929,7 +1066,7 @@ mod tests {
 
     #[test]
     fn not_present_empty() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
 
         let (_, err) = IgnoreBuilder::new().build().add_child(td.path());
         assert!(err.is_none());
@@ -939,7 +1076,7 @@ mod tests {
     fn stops_at_git_dir() {
         // This tests that .gitignore files beyond a .git barrier aren't
         // matched, but .ignore files are.
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         mkdirp(td.path().join(".git"));
         mkdirp(td.path().join("foo/.git"));
         wfile(td.path().join(".gitignore"), "foo");
@@ -960,7 +1097,7 @@ mod tests {
 
     #[test]
     fn absolute_parent() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         mkdirp(td.path().join(".git"));
         mkdirp(td.path().join("foo"));
         wfile(td.path().join(".gitignore"), "bar");
@@ -983,7 +1120,7 @@ mod tests {
 
     #[test]
     fn absolute_parent_anchored() {
-        let td = tmpdir("ignore-test-");
+        let td = tmpdir();
         mkdirp(td.path().join(".git"));
         mkdirp(td.path().join("src/llvm"));
         wfile(td.path().join(".gitignore"), "/llvm/\nfoo");
@@ -1000,4 +1137,49 @@ mod tests {
         assert!(ig2.matched("foo", false).is_ignore());
         assert!(ig2.matched("src/foo", false).is_ignore());
     }
+
+    #[test]
+    fn git_info_exclude_in_linked_worktree() {
+        let td = tmpdir();
+        let git_dir = td.path().join(".git");
+        mkdirp(git_dir.join("info"));
+        wfile(git_dir.join("info/exclude"), "ignore_me");
+        mkdirp(git_dir.join("worktrees/linked-worktree"));
+        let commondir_path =
+            || git_dir.join("worktrees/linked-worktree/commondir");
+        mkdirp(td.path().join("linked-worktree"));
+        let worktree_git_dir_abs = format!(
+            "gitdir: {}",
+            git_dir.join("worktrees/linked-worktree").to_str().unwrap(),
+        );
+        wfile(td.path().join("linked-worktree/.git"), &worktree_git_dir_abs);
+
+        // relative commondir
+        wfile(commondir_path(), "../..");
+        let ib = IgnoreBuilder::new().build();
+        let (ignore, err) = ib.add_child(td.path().join("linked-worktree"));
+        assert!(err.is_none());
+        assert!(ignore.matched("ignore_me", false).is_ignore());
+
+        // absolute commondir
+        wfile(commondir_path(), git_dir.to_str().unwrap());
+        let (ignore, err) = ib.add_child(td.path().join("linked-worktree"));
+        assert!(err.is_none());
+        assert!(ignore.matched("ignore_me", false).is_ignore());
+
+        // missing commondir file
+        assert!(fs::remove_file(commondir_path()).is_ok());
+        let (_, err) = ib.add_child(td.path().join("linked-worktree"));
+        // We squash the error in this case, because it occurs in repositories
+        // that are not linked worktrees but have submodules.
+        assert!(err.is_none());
+
+        wfile(td.path().join("linked-worktree/.git"), "garbage");
+        let (_, err) = ib.add_child(td.path().join("linked-worktree"));
+        assert!(err.is_none());
+
+        wfile(td.path().join("linked-worktree/.git"), "gitdir: garbage");
+        let (_, err) = ib.add_child(td.path().join("linked-worktree"));
+        assert!(err.is_none());
+    }
 }
